﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Threading;

using System.Runtime.CompilerServices;

namespace AsyncCustomAwaiters {

	class ThreadedWorker {
		private volatile bool isCompleted = false;
		private volatile Action continuation = null;

		private Thread thread;

		public ThreadedWorker() {
			thread = new Thread(() => {
				Console.WriteLine("*** ThreadedWorker in thread {0}: Starting sleeping ...", Thread.CurrentThread.ManagedThreadId);
				Thread.Sleep(1000);
				Console.WriteLine();
				Console.WriteLine("*** ThreadedWorker in thread {0}: Done sleeping, going to Completed state ...", Thread.CurrentThread.ManagedThreadId);
				isCompleted = true;

				var c = continuation;
				Console.WriteLine("*** ThreadedWorker in thread {0}: Continuation content: {1}", Thread.CurrentThread.ManagedThreadId, c);
				if (c != null) {
					Console.WriteLine("*** ThreadedWorker in thread {0}: Executing continuation ...", Thread.CurrentThread.ManagedThreadId, c);
					c();
				}
				Console.WriteLine("*** ThreadedWorker in thread {0}: Dying ...", Thread.CurrentThread.ManagedThreadId, c);
			});

			Console.WriteLine("* ThreadedWorker.ctor in thread {0}: created new worker in thread {1}", Thread.CurrentThread.ManagedThreadId, thread.ManagedThreadId);

			thread.Start();
		}

		public Awaiter GetAwaiter() {
			Console.WriteLine("* ThreadedWorker.GetAwaiter() in thread {0}: returning awaiter for thread {1}", Thread.CurrentThread.ManagedThreadId, thread.ManagedThreadId);
			return new Awaiter(this);
		}

		public struct Awaiter : INotifyCompletion {
			private ThreadedWorker worker;

			public Awaiter(ThreadedWorker worker) {
				this.worker = worker;
			}

			public bool IsCompleted { 
				get {
					var result = worker.isCompleted;
					Console.WriteLine("== ThreadedWorker({2}).Awaiter.IsCompleted called in thread {0} -> returing {1}", Thread.CurrentThread.ManagedThreadId, result, worker.thread.ManagedThreadId);
					return result;
				}
			}

			// Caller expects exceptions thrown from this method if task failed.
			//		TaskCanceledException: The task was canceled.
			//		Exception: The task completed in a Faulted state.
			// This method can have also any returnig type instead of "void" - for "future"-like custom task awaiters.
			public void GetResult() {
				Console.WriteLine("== ThreadedWorker({1}).Awaiter.GetResult() called in thread {0}", Thread.CurrentThread.ManagedThreadId, worker.thread.ManagedThreadId);
			}

			public void OnCompleted(Action continuation) {
				Console.WriteLine("== ThreadedWorker({2}).Awaiter.OnCompleted({1}) called in thread {0}", Thread.CurrentThread.ManagedThreadId, continuation, worker.thread.ManagedThreadId);
				worker.continuation = continuation;
			}
		}
	}

	class Program {
		static async void RunTestAsync() {
			Console.WriteLine("Program.RunTestAsync in thread {0}: Initial part 1 running ...", Thread.CurrentThread.ManagedThreadId);
			await new ThreadedWorker();
			Console.WriteLine("Program.RunTestAsync in thread {0}: Part 2 after 1st await ...", Thread.CurrentThread.ManagedThreadId);
			await new ThreadedWorker();
			Console.WriteLine("Program.RunTestAsync in thread {0}: Part 3 after 2st await ...", Thread.CurrentThread.ManagedThreadId);
		}

		static void Main(string[] args) {
			Console.WriteLine("!!! Program.Main in thread {0}: Started ...", Thread.CurrentThread.ManagedThreadId);

			RunTestAsync();

			Console.WriteLine("!!! Program.Main in thread {0}: Ending ...", Thread.CurrentThread.ManagedThreadId);
		}
	}
}
